import os
from pathlib import Path
import numpy as np


def check_rm(path):
    """
    Check and delete file

    :param path: path string of target file
    """
    if os.path.isfile(path):
        print(f"removing {path}")
        os.remove(path)


def file_exists(path):
    if os.path.isfile(path):
        return True
    else:
        raise FileNotFoundError(f"{path} not exist")


def is_empty(path):
    """
    Check if file is empty. If the file does not exist, then create one.

    :param path: path string of target file
    :returns: boolean operator
    """
    if os.path.isfile(path):
        return os.stat(path).st_size == 0
    else:
        Path(path).touch()

def efficient_log(data):
    "keep NA, remove zeros"
    return np.where(data == 0, -10, np.log10(data))


from os.path import join
from typing import Union, List, Tuple
from pathlib import Path
import re
import warnings

import pandas as pd
from netCDF4 import Dataset
import xarray as xr
from scipy.ndimage import binary_dilation
import numpy as np

from .utils import file_exists
from .chem import Chemistry
from .array import GriddedData


class GenieModel(object):
    """
    GenieModel is the interface to cGENIE output
    """

    def __init__(self, model_path: Union[str, List, Tuple], gemflag=None):
        """
        Initialise a GenieModel object with a path to cGENIE output directory

        -------
        Example
        >>> from cgeniepy.model import GenieModel
        >>> model = GenieModel("path_to_GENIE_output")

        ## specify cgenie component
        >>> model = GenieModel("path_to_GENIE_output", gemflag=["biogem"])
        """

        ## check if model_path is a valid directory
        if isinstance(model_path, (list, tuple)):
            self.is_ensemble = True
            for path in model_path:
                if not Path(path).is_dir():
                    raise ValueError(f"{path} is not a valid directory")
        elif isinstance(model_path, str):
            self.is_ensemble = False
            if not Path(model_path).is_dir():
                raise ValueError(f"{model_path} is not a valid directory")

        if not gemflag:
            warnings.warn("No gemflag is provided, use default gemflags: [biogem]")
            self.gemflag = ["biogem"]
        else:
            self.gemflag = gemflag
            ## if gemflag is a string, convert it to a list
            if isinstance(gemflag, str):
                self.gemflag = [gemflag]

        self.model_path = model_path
        self.ncvar_dict = self._ncvar_dict()

    def _model_ncpath(self, gem="ecogem", dim="2d"):
        """
        get the path of model NetCDF output of target gem and dimension
        if the model is an ensemble, return a list of paths
        """

        nc_file = f"fields_{gem}_{dim}.nc"

        if not self.is_ensemble:
            model_path = self.model_path
            nc_path = join(model_path, gem, nc_file)
            if file_exists(nc_path):
                return nc_path
        else:
            nc_paths = []
            for path in self.model_path:
                model_path = path
                nc_path = join(model_path, gem, nc_file)
                if file_exists(nc_path):
                    nc_paths.append(nc_path)
            ## make `nc_paths` hashable
            nc_paths = tuple(nc_paths)
            return nc_paths

    def _ncvar_dict(self) -> dict:
        """
        Return all available variables and related biogem/ecogem (2d or 3d)
        NetCDF path for each model
        """

        ## initialise a dictionary, key: model_path, value: ncvar_list
        var_path = {}

        for gem in self.gemflag:
            for dim in ["2d", "3d"]:
                ## ignore ecogem 3d which is not available now (Oct 2023)
                if gem == "ecogem" and dim == "3d":
                    continue

                if self.is_ensemble:
                    ## use one model to get the nc_path
                    nc_path = self._model_ncpath(gem, dim)[0]
                    all_nc_path = self._model_ncpath(gem, dim)
                else:
                    nc_path = self._model_ncpath(gem, dim)

                nc = Dataset(nc_path, "r")
                ncvar_list = list(nc.variables.keys())
                nc.close()

                ## add key, value to the dictionary
                if self.is_ensemble:
                    var_path[all_nc_path] = ncvar_list
                else:
                    var_path[nc_path] = ncvar_list

        return var_path

    def _lookup_ncpath(self, var):
        """
        find the netcdf path of the target variable

        if more than one model satisfies the condition,
        return all model's NetCDF paths
        """
        for path, value_list in self.ncvar_dict.items():
            if var in value_list:
                return path
        raise ValueError(f"variable {var} not found in the ncvar_dict")

    def _open_nc(self, nc_path):
        "Use xarray to open netcdf file"
        ## if path is a list of paths
        if not isinstance(nc_path, (list, tuple)):
            ## `open_dataset` to lazy load the data
            return xr.open_dataset(nc_path)
        else:
            datasets = [xr.open_dataset(file) for file in nc_path]
            combined_ds = xr.concat(datasets, "model")
            combined_ds.coords["model"] = self.model_path
            return combined_ds

    def get_var(self, var: Union[str, List, Tuple], attrs=None, mutable=True):
        """
        Get the data of target variable.
        A list of variables is supported as well.

        :param var: the name of target variable
        :param unit: the unit of target variable, usually provided in the model output
        """
        self.target_var = var

        ## if varstr is a string
        if isinstance(self.target_var, str):
            ## find the path to the netcdf file
            path2nc = self._lookup_ncpath(var=self.target_var)
            ## open the netcdf file
            array = self._open_nc(path2nc)[self.target_var]
        ## if varstr is a list/tuple of strings
        ## concat the data along the "variable" dimension
        elif isinstance(self.target_var, (list, tuple)):
            array_container = []
            for v in self.target_var:
                path2nc = self._lookup_ncpath(var=v)
                single_array = self._open_nc(path2nc)[v]
                array_container.append(single_array)

            array = xr.concat(array_container, "variable")
            array.name = "ensemble_variable"

        ## initialise GriddedData object
        if not attrs:
            return GriddedData(array, mutable=mutable, attrs=array.attrs)
        else:
            return GriddedData(array, mutable=mutable, attrs=attrs)
            

    def tsvar_list(self):
        """list all files biogem timeseries files

        If the model is an ensemble, it assumes that all models share the same
        file/data structure and only return the first model's timeseries list
        """

        if self.is_ensemble:
            biogem_path = join(self.model_path[0], "biogem")
        else:
            biogem_path = join(self.model_path, "biogem")

        ## return all files in the directory
        all_bg_files = Path(biogem_path)

        ## find out files with .res extension
        tsvar_list = [f for f in all_bg_files.glob("*.res")]

        return tsvar_list

    def get_ts(self, var: str):
        """
        read in time series output of GENIE

        :param filename: the name of the time series file
        :return: a pandas DataFrame
        """

        if not self.is_ensemble:
            filename = f"biogem_series_{var}.res"            
            f = join(self.model_path, "biogem", filename)
            if not file_exists(f):
                raise ValueError(f"{f} does not exist")

            # Read the text file into a DataFrame
            with open(f, "r") as file:
                lines = file.readlines()

            # Remove " %" from the header and split it using " / "
            # not '/' which is also included in the isotope unit (e.g. 'd13C o/oo')
            header = [col.strip().lstrip("% ") for col in lines[0].split(" / ")]
            data = [line.split() for line in lines[1:]]
            df = pd.DataFrame(data, columns=header)

            ## convert to numeric
            df = df.apply(pd.to_numeric, errors="coerce")

            ## add a column of model name
            df["model"] = self.model_path.split("/")[-1]
            return df
        else:
            ## concatenate all models' data into one data frame
            df_list = []

            for path in self.model_path:
                ## do the same thing as above
                filename = f"biogem_series_{var}.res"
                f = join(path, "biogem", filename)
                if not file_exists(f):
                    raise ValueError(f"{f} does not exist")

                ## read text file into a DataFrame
                with open(f, "r") as file:
                    lines = file.readlines()

                ## Remove " %" from the header and split it using " / "
                header = [col.strip().lstrip("%") for col in lines[0].split(" / ")]
                data = [line.split() for line in lines[1:]]

                df = pd.DataFrame(data, columns=header)

                ## convert to numeric
                df = df.apply(pd.to_numeric, errors="coerce")

                df["model"] = path.split("/")[-1]
                df_list.append(df)

            ## concatenate by row
            all_df = pd.concat(df_list, axis=0)

            return all_df
        
    def get_diag_avg(self, filename):
        "print out the summary of a diagnostic file"
        f = join(self.model_path, "biogem", filename)
        with open(f, 'r') as f:
            text = f.read()

        # Split data by lines, remove empty lines
        lines = [line.strip() for line in text.splitlines() if line.strip()]

        # Initialize an empty dictionary to store all data
        variable = []
        value1 = []
        value2 = []

        # Loop through each line
        for line in lines:
        # Split line by separator (colon)
            try:
                key, value = line.split(':', 1)
            except ValueError:  # Ignore lines without a colon separator
                continue

            # Remove leading/trailing whitespaces from key and value
            key = key.strip()
            value = value.strip()

            if '<->' in value:
                v1 = value.split('<->')[0].strip()
                v2 = value.split('<->')[1].strip()
            else:
                v1 = value
                v2 = None

            key = re.sub(r"\s*\.+$", "", key)

            variable.append(key)
            value1.append(v1)
            value2.append(v2)
    
        
        ## get the model year and make it a separate column
        year = value1[0]

        ## convert data to data frame
        df = pd.DataFrame({'year': year,
                           'variable': variable, 
                           'value1': value1, 
                           'value2': value2})
        
        ## remove the duplicated first row
        df = df.iloc[1:]

        return df

    def grid_mask(self):
        """
        cGENIE continent mask array (by setting zero values),
        either calculated from existing data or using biogem.grid_mask
        """
        try:
            grid_mask = self.get_var("grid_mask").data
            return grid_mask
        except ValueError:
            print("grid_mask not found!")

    def grid_category(self):
        """an alogirthm to define surface grid catogories depending on the land-sea mask
        0: coastal sea
        1: land
        2: open ocean
        """

        is_sea = np.isnan(self.grid_mask())
        land_around = binary_dilation(is_sea)
        grid_catogories = xr.where(land_around, np.logical_and(land_around, is_sea), 2)

        return grid_catogories

    def grid_area(self):
        "grid area array in m2"
        ## read grid_area from biogem
        print("grid area returned in the unit of 'm2'")
        return self.get_var("grid_area")
    
    def grid_mask_3d(self):
        return self.get_var("grid_mask_3d")
    
    def grid_topo(self):
        return self.get_var("grid_topo")
    
    def grid_zt_edges(self):
        return self.get_var("zt_edges")
    
    def grid_lat_edges(self):
        return self.get_var("lat_edges")
    
    def grid_lon_edges(self):
        return self.get_var("lon_edges")
    
    def grid_zt_depths(self):
        return self.get_var("grid_dD")
    
    def grid_volume(self):
        "grid volume array (3d) in m3"
        grid_area = self.grid_area()  ## m2
        depth = self.grid_zt_depths() ## m
        ocn_mask = self.grid_mask_3d()
        try:
            grid_volume = depth * grid_area * ocn_mask 
            grid_volume.data.attrs["units"] = "m$^{3}$"
            grid_volume.data.attrs["long_name"] = "grid volume"
            print("grid volume calculated in the unit of 'm3'")
        except ValueError:
            print(
                "Depth array not found! Please ensure 3d data is exported in the model!"
            )

        return grid_volume



import pathlib
from cgeniepy.skill import DFComparison
import geopandas as gpd
from shapely.geometry import Point
import pandas as pd

from io import StringIO

from cgeniepy.grid import Interporaltor

from .plot import ScatterDataVis
from .grid import GridOperation
import cgeniepy.array as ca


class ScatterData:
    """ScatterData is a class to store non-gridded data, often seen in the observations
    """

    def __init__(self, data, *args, **kwargs):
        """
        Initialize a ScatterData object.

        Parameters:
        data: The path to the file or the data.
        coord_cols (dict): A dictionary specifying the coordinate columns.
        """
        ## if already a dataframe
        if isinstance(data, pd.DataFrame):
            self.data = data

        ## if a file path then read the file into a dataframe
        if isinstance(data, str):
            if data.endswith(".tab"):
                data = self.parse_tab_file(data)
                self.data= pd.read_csv(StringIO(data), *args, **kwargs)
            elif data.endswith("xlsx"):
                self.data = pd.read_excel(data, *args, **kwargs)                
            else:
                self.data = pd.read_csv(data, *args, **kwargs)

        ## if the index is already set in the data, then set the coordinates
        if not isinstance(self.data.index, pd.core.indexes.range.RangeIndex):
            self.index= list(self.data.index.names)
            GridOperation().set_coordinates(obj=self, index=self.index)

    def set_index(self, index):
        "set the index of the dataframe"
        self.data.set_index(index, inplace=True)
        self.index= index

    def parse_tab_file(self, filename, begin_cmt = '/*', end_cmt = '*/'):
        """
        Read a tab-delimited file and return a pandas DataFrame

        This function is optimised for pangea-format data.
        """
        lines = []
        in_comment = False

        with open(filename) as f:
            for line in f:
                if line.strip().startswith(begin_cmt):
                    in_comment = True
                    continue

                if line.strip().startswith(end_cmt):
                    in_comment = False
                    continue

                if not in_comment:
                    lines.append(line.rstrip('\n'))

        data = '\n'.join(lines)
        return data

    def __getitem__(self, item):
        return self.data[item]
    
    def _check_cols(self, cols):
        """
        Check if the columns are present in the dataframe.

        Args:
            cols (list): List of column names to check.

        Raises:
            ValueError: If any of the columns are not found in the dataframe.
        """
        for col in cols:
            if col not in self.data.columns:
                raise ValueError(f"{col} not found in the dataframe")

    def detect_basin(self):
        "use point-in-polygon strategy to detect modern ocean basin according to lon/lat column"
        
        def sub_detect_basin(lon, lat):
            p = Point(lon, lat)
            ocean_name = oceans[oceans.contains(p)].Oceans.values
            if ocean_name.size > 0:
                return ocean_name[0]
            else:
                return ""
        file_path = pathlib.Path(__file__).parent.parent / "data/oceans/oceans.shp"
        oceans = gpd.read_file(file_path)
        
        result = self.data.reset_index().apply(lambda row: sub_detect_basin(row[self.lon], row[self.lat]), axis=1)

        ## add to self.data
        tmp_data = self.data.reset_index()
        tmp_data['basin'] = result
        tmp_index = self.index
        y = ScatterData(tmp_data)
        y.set_index(tmp_index)
        return y


    def to_xarray(self):
        "convert to xarray dataset"
        ## set the coordinate using the data from self.coordinates         
        return self.data.to_xarray()

    def to_GriddedData(self, var):
        "convert to gridded data"
        output = ca.GriddedData(self.to_xarray()[var])
        return output

    def interpolate(self, var):
        ## a tuple of coordinate arrays
        coords =  tuple([self.data.reset_index(inplace=False)[dim].values for dim in self.index])
        ## array values
        values = self.data[var].values
        dims = self.index
        output= Interporaltor(dims, coords, values, 200, 'ir-linear')
        return output


    def to_genie(
        self,
        low_threshold=None,
        new_low_bound=None,
        high_threshold=None,
        new_high_bound=None,
    ):
        """
        Regrid a dataframe within certain format to cGENIE grids

        Input: A dataframe with three necessary columns: "Latitude", "Longitude", "Observation".
        The other columns will be unselected. Note that observation column is the data you want to process.
        Rename dataframe by using command: `df = df.rename({"old name": "new name"}, axis='columns')`

        Output:
        A 36x36 2D array.

        Optional parameters:
        low_threshold/new_low_bound: change the data that is smaller than low_threshold to new_low_bound.
        For example, covert any data < 0.03 to 0. Work same to high_threshold/new_high_bound.
        """

        df = self.data.copy()

        # subset
        df = df[["Longitude", "Latitude", "Observation"]]

        # drop NAN
        df = df.dropna(axis="rows", how="any")

        # regrid coordinate
        df.loc[:, "Latitude"] = df.loc[:, "Latitude"].apply(regrid_lat)
        df.loc[:, "Longitude"] = df.loc[:, "Longitude"].apply(regrid_lon)

        # group and aggregate (still in long format)
        df_agg = df.groupby(["Longitude", "Latitude"]).agg("mean")

        # Create a all-nan data frame with full cGENIE grids
        lat = GENIE_lat()
        lon = normal_lon()
        data = np.zeros([36 * 36])
        index = pd.MultiIndex.from_product([lon, lat], names=["Longitude", "Latitude"])
        df_genie = pd.DataFrame(data, index=index, columns=["Observation"])
        df_genie["Observation"] = df_genie["Observation"].replace(0, np.nan)

        # transform the tuple elements in multindex to lists
        df_genie_index_list = [list(item) for item in df_genie.index.values]
        df_agg_index_list = [list(item) for item in df_agg.index.values]

        # copy aggregated dataframe to full-grid dataframe, one by one
        # if the returned dataframe is full of NA, it's very likely from this step
        for i in df_genie_index_list:

            longitude = i[0]
            latitude = i[1]

            if i in df_agg_index_list:
                df_genie.loc[longitude, latitude] = df_agg.loc[longitude, latitude]
            else:
                df_genie.loc[longitude, latitude] = np.nan

        # filter data if necessary
        if (low_threshold is not None) and (new_low_bound is not None):
            df_genie.loc[
                df_genie.Observation < low_threshold, "Observation"
            ] = new_low_bound

        if (high_threshold is not None) and (new_high_bound is not None):
            df_genie.loc[
                df_genie.Observation > high_threshold, "Observation"
            ] = new_high_bound

        # long -> wide data format
        df_genie_wide = df_genie.pivot_table(
            values="Observation", index="Latitude", columns="Longitude", dropna=False
        )

        return df_genie_wide
        
    def drop_na(self, *args, **kwargs):
        "drop rows with NA values"
        self.data = self.data.dropna(*args, **kwargs)

    def to_ScatterDataVis(self):
        "convert to ScatterDataVis"
        return ScatterDataVis(self)

    def plot(self, var, *args, **kwargs):
        "plot the data"
        return self.to_ScatterDataVis().plot(var, *args, **kwargs)

    def compare(self, var1, var2):
        "compare two ScatterData objects"
        return DFComparison(self.data, var1, var2)




import pandas as pd
import numpy as np
import re

from .model import GenieModel

class EcoModel(GenieModel):
    """
    EcoModel is an subclass of GenieModel, as EcoGENIE to cGENIE

    It facilitates the access of ecophysiological parameters and plankton variables

    Initialise a EcoModel object with a path to cGENIE output directory

    -------
    Example
    >>> from cgeniepy.ecology import EcoModel
    >>> model = EcoModel("path_to_GENIE_output")
    """

    def __init__(self, *args, **kwargs):
        
        super().__init__(*args, **kwargs)        

        ## get the list of plankton variables
        ecogem2d_path = self._model_ncpath('ecogem', '2d')            
        self.eco_varlist = self.ncvar_dict[ecogem2d_path]
        
        pattern = r'eco2D_Plankton_C_\d+'
        plank_vars = [var for var in self.eco_varlist if re.match(pattern, var)]
        self.plank_n = len(plank_vars)

        if self.is_ensemble:
            sub_ecopars = self.eco_pars().iloc[range(self.plank_n),]
        else:
            sub_ecopars = self.eco_pars()

        photo_par = sub_ecopars['vmax_C'].to_numpy()
        graz_par = sub_ecopars['max_graz_C'].to_numpy()

        autotrophy = (photo_par != 0)
        heterotrophy = (graz_par != 0)
        mixotrophy = np.logical_and(autotrophy, heterotrophy)

        phyto_idx = np.where(np.logical_and(autotrophy, ~mixotrophy))[0]
        zoo_idx = np.where(np.logical_and(heterotrophy, ~mixotrophy))[0]
        mixo_idx = np.where(mixotrophy)[0]

        ## convert to ECOGENIE format
        self.phyto_indices = [f'{i+1:03d}' for i in phyto_idx]
        self.zoo_indices = [f'{i+1:03d}' for i in zoo_idx]
        self.mixo_indices = [f'{i+1:03d}' for i in mixo_idx]
        
        ## some basic information (number of PFTs)
        self.phyto_n = len(self.phyto_indices)
        self.zoo_n = len(self.zoo_indices)
        self.mixo_n = len(self.mixo_indices)

        ## check everything is correct
        if (self.phyto_n + self.zoo_n + self.mixo_n != self.plank_n):
            raise ValueError("The number of PFTs does not match the number of plankton variables")


    def eco_pars(self):
        """
        Get all the ecophysiological parameters used in the model

        :return: a pandas DataFrame object
        """
        if self.is_ensemble:
            df_list = []
            for path in self.model_path:
                path = f"{path}/ecogem/Plankton_params.txt"
                df = pd.read_fwf(path)
                df_list.append(df)
            df_all = pd.concat(df_list)
            return df_all
        else:
            path = f"{self.model_path}/ecogem/Plankton_params.txt"
            df = pd.read_fwf(path)
            return df
            

    def get_pft(self, pft_index, prefix="Plankton", element="C"):
        """
        a variant of GenieModel's `get_var`, to select plankton functional type

        :param pft_index: the index of plankton functional type
        :param prefix: 'Plankton' or 'Export'
        :param element: 'C', 'Fe', 'P', 'Si', 'N'

        :return: a GenieArray object

        ----------------------------------------
        Example:

        >>> from cgeniepy.ecology import EcoModel

        >>> ### initialise a model object
        >>> model = EcoModel("path_to_GENIE_output")
        
        >>> ### get PFT-1 carbon biomass data
        >>> model.get_pft(1, "Plankton", "C")

        >>> ### get a list of PFT carbon biomass data
        >>> model.get_pft([1, 2, 3], "Plankton", "C")

        >>> ### get all phytoplankton carbon biomass data
        >>> model.get_pft('phyto', "Plankton", "C")

        >>> ### get all zooplankton carbon biomass data
        >>> model.get_pft('zoo', "Plankton", "C")

        >>> ### get all plankton carbon biomass data
        >>> model.get_pft('all', "Plankton", "C")
        """

        ### >>> preprocess the prefix
        ## if prefix is "Biomass", replace it with "Plankton"
        if prefix == "Biomass": prefix = "Plankton"        
        ## always capitalize the prefix
        ## prefix = prefix.capitalize()

        ### >>> process the pft_index
        ## if pft_index is "phyto", replace it with phyto_indices
        if pft_index == "phyto": pft_index = self.phyto_indices
        if pft_index == "zoo": pft_index = self.zoo_indices
        if pft_index == "mixo": pft_index = self.mixo_indices

        ## >>> construct variable name using pft_index, prefix and element
        if isinstance(pft_index, (int, str)):
            fullstring = f"eco2D_{prefix}_{element}_{pft_index:03}"
        elif isinstance(pft_index, (list, tuple)):
            fullstring = []
            for i in pft_index:
                fullstring.append(f"eco2D_{prefix}_{element}_{i:03}")
        else:
            raise ValueError("pft_index must be an integer or a list/tuple of integers")

        ## modify the fullstring if pft_index is "all"
        ## eco2D_Plankton_C_Total/eco2D_Plankton_Chl_Total
        if pft_index == "all":
            fullstring = f"eco2D_{prefix}_{element}_Total"
        
        return self.get_var(fullstring)



import numpy as np
from scipy.spatial import distance
import matplotlib.pyplot as plt

class ArrComparison:
    
    "quantitatively compare similarity metrics between two 2D arrays (model and observation)"

    def __init__(self, model, observation):
        """
        
        :param: model: array-like data
        :param: observation: array-like data

        --------
        Example:
        --------
        
        from cgeniepy.skill import ModelSkill

        ## generate 2d data
        import numpy as np
        x = np.random.rand(36, 36)
        y = np.random.rand(36, 36)

        ## calculate skill score
        ms = ModelSkill(x, y)
        ms.mscore()
        """
        self.model = model
        self.data = observation

        ## check data type, if not float, convert to float
        if self.model.dtype != np.float64:
            self.model = self.model.astype(np.float64)
        if self.data.dtype != np.float64:
            self.data = self.data.astype(np.float64)
        

    def safe_unveil(self, array):
        "get pure array from a numpy masked array object"
        if array.__class__ != np.ma.core.MaskedArray:
            return array
        else:
            return array.filled(np.nan)

    def intersect_index(self, verbose=False):
        """ Return the index where corresponding values are not
        nan in both input arrays. One then can filter the array
        by the output boolean array.
        """

        # If both are not NaN, return it
        array1 = self.safe_unveil(self.model)
        array2 = self.safe_unveil(self.data)

        indx_array = np.logical_and(~np.isnan(array1), ~np.isnan(array2))

        if verbose is True:
            num = indx_array.flatten()[indx_array.flatten() == True].shape[0]
            print("Summary: {} elements simultaneously exist.".format(num))

        return indx_array

    def mscore(self):
        """
        Calculate skill metric M-score. See more in the paper Watterson, I. G. (1996)
        """

        # get common set
        indx = self.intersect_index()
        sub_data1 = self.model[indx]
        sub_data2 = self.data[indx]

        # calculate M-score
        mse = np.square(np.subtract(sub_data1, sub_data2)).mean()
        v1 = sub_data1.var()
        v2 = sub_data2.var()
        g1 = sub_data1.mean()
        g2 = sub_data2.mean()

        mscore = (2 / np.pi) * np.arcsin(1 - mse / (v1 + v2 + np.square(g1 - g2)))

        return mscore

    def pearson_r(self):
        """
        calculate pearson correlation coefficient for 2D array
        """
        # get common set
        indx = self.intersect_index()
        sub_data1 = self.model[indx]
        sub_data2 = self.data[indx]

        # correlation matrix
        corr_mat = np.corrcoef(sub_data1, sub_data2)
        corr = corr_mat[0, 1]

        return corr

    def cos_similarity(self):
        """
        Cosine similarity that ignores the magnitude of data, but focuses on the direction
        """
        # get common set
        indx = self.intersect_index()
        sub_data1 = self.model[indx].ravel()
        sub_data2 = self.data[indx].ravel()

        if sub_data1.mean() != 0 and sub_data2.mean() != 0:
            cos_sim = 1 - distance.cosine(sub_data1.flatten(), sub_data2.flatten())
        else:
            cos_sim = np.nan

        return cos_sim

    def rmse(self):
        """
        Root Mean Sqaure Error (rmse, or rmsd)
        """
        data1 = self.model
        data2 = self.data

        error_2d = data1 - data2
        error_1d = error_2d.ravel()[~np.isnan(error_2d.ravel())]
        rmse = np.sqrt(np.square(error_1d).mean())

        return rmse

    def nrmse(self):
        "Normalised RMSE facilitating the comparison between datasets or models with different scales"

        norm_rmse = self.rmse()/np.nanmean(self.model)
        return norm_rmse

    def crmse(self):
        """
        centred Root Mean Sqaure Error (rmse, or rmsd). See Talor, K. E. (2001) JGR
        """
        # select data
        indx = self.intersect_index()
        sub_data1 = self.model[indx].ravel()
        sub_data2 = self.data[indx].ravel()

        # calculate std
        sigma1 = np.std(sub_data1)
        sigma2 = np.std(sub_data2)

        # pearson correlation
        corr_mat = np.corrcoef(sub_data1, sub_data2)
        corr = corr_mat[0, 1]

        # central rmse
        crmse = sigma1**2 + sigma2**2 - 2 * sigma1 * sigma2 * corr

        return crmse

    def plot(self):
        ## a x-y plot
        fig, ax = plt.subplots()
        ax.scatter(self.model, self.data)
        ax.set_xlabel("Model")
        ax.set_ylabel("Observation")
        ax.set_title("Model vs Observation")

    
class DFComparison(ArrComparison):

    def __init__(self, df, model_col, observation_col):
        self.model = df[model_col].to_numpy()
        self.data = df[observation_col].to_numpy()
        super().__init__(self.model, self.data)

 
# class TaylorDiagram(object):
#     """
#     Taylor diagram.
#     Plot model standard deviation and correlation to reference data in a single-quadrant polar plot,
#     with r=stddev and theta=arccos(correlation).

#     modified from Yannick Copin, https://gist.github.com/ycopin/3342888
#     reference: https://matplotlib.org/stable/gallery/axisartist/demo_floating_axes.html
#     """

#     def __init__(
#             self,
#             fig=None,
#             figscale=1,
#             subplot=111,
#             xmax=None,
#             tmax=np.pi / 2,
#             ylabel="Standard Deviation",
#             rotation=None,
#     ):

#         """
#         Set up Taylor diagram axes, i.e. single quadrant polar
#         plot, using `mpl_toolkits.axisartist.floating_axes`.

#         Parameters:

#         * fig: input Figure or None
#         * subplot: subplot definition
#         * xmax: the length of radius, xmax can be 1.5* reference std
#         """

#         # --------------- tickers --------------------------
#         # Correlation labels (if half round)
#         cor_label = np.array([0, 0.2, 0.4, 0.6, 0.8, 0.9, 0.95, 0.99, 1])

#         # add the negative ticks if more than half round
#         excess_theta = tmax - np.pi / 2
#         if excess_theta > 0:
#             cor_label = np.concatenate((-cor_label[:0:-1], cor_label))

#         # convert to radian
#         rad = np.arccos(cor_label)
#         # tick location
#         gl = gf.FixedLocator(rad)
#         # tick formatting: bind radian and correlation coefficient
#         tf = gf.DictFormatter(dict(zip(rad, map(str, cor_label))))

#         # --------------- coordinate -----------------------
#         # Standard deviation axis extent (in units of reference stddev)
#         # xmin must be 0, which is the centre of round

#         self.xmin = 0
#         self.xmax = xmax
#         self.tmax = tmax

#         # ------- curvilinear coordinate definition -------
#         # use built-in polar transformation (i.e., from theta and r to x and y)
#         tr = PolarAxes.PolarTransform()
#         ghelper = fa.GridHelperCurveLinear(
#             tr,
#             extremes=(0, self.tmax, self.xmin, self.xmax),
#             grid_locator1=gl,
#             tick_formatter1=tf,
#         )

#         # ------- create floating axis -------
#         if fig is None:
#             fig_height = 4.5 * figscale
#             fig_width = fig_height * (1 + np.sin(excess_theta))
#             fig = plt.figure(figsize=(fig_width, fig_height), dpi=100)

#         ax = fa.FloatingSubplot(fig, subplot, grid_helper=ghelper)
#         fig.add_subplot(ax)

#         # Adjust axes
#         # Angle axis
#         ax.axis["top"].label.set_text("Correlation")
#         ax.axis["top"].toggle(ticklabels=True, label=True)
#         # inverse the direction
#         ax.axis["top"].set_axis_direction("bottom")
#         ax.axis["top"].major_ticklabels.set_axis_direction("top")
#         ax.axis["top"].label.set_axis_direction("top")

#         # X axis
#         ax.axis["left"].set_axis_direction("bottom")

#         # Y axis direction & label
#         ax.axis["right"].toggle(all=True)
#         ax.axis["right"].label.set_text(ylabel)
#         ax.axis["right"].set_axis_direction("top")
#         # ticklabel direction
#         ax.axis["right"].major_ticklabels.set_axis_direction("left")

#         ax.axis["bottom"].set_visible(False)

#         # ------- Set instance attribute ----------
#         self.fig = fig
#         # Graphical axes
#         self._ax = ax
#         # grid line
#         self._ax.grid(True, zorder=0, linestyle="--")
#         # aspect ratio
#         self._ax.set_aspect(1)
#         # A parasite axes for further plotting data
#         self.ax = ax.get_aux_axes(tr)
#         # Collect sample points for latter use (e.g. legend)
#         self.samplePoints = []

#     def add_ref(self, refstd, reflabel="Observation", linestyle="-", color="k"):
#         """add a reference point"""
#         self.refstd = refstd
#         # Add reference point
#         # slightly higher than 0 so star can be fully seen
#         l = self.ax.plot(0.01, self.refstd, "k*", ls="", ms=10)
#         # xy for the point, xytext for the text (the coordinates are
#         # defined in xycoords and textcoords, respectively)
#         self.ax.annotate(
#             reflabel,
#             xy=(0.01, self.refstd),
#             xycoords="data",
#             xytext=(-25, -30),
#             textcoords="offset points",
#         )
#         # add stddev contour
#         t = np.linspace(0, self.tmax)
#         r = np.zeros_like(t) + self.refstd
#         self.ax.plot(t, r, linestyle=linestyle, color=color)
#         self.samplePoints.append(l)

#     def add_scatter(self, stddev, corrcoef, *args, **kwargs):
#         """
#         Add sample (*stddev*, *corrcoeff*) to the Taylor
#         diagram. *args* and *kwargs* are directly propagated to the
#         `Figure.plot` command.
#         """

#         l = self.ax.scatter(
#             np.arccos(corrcoef), stddev, *args, **kwargs
#         )  # (theta, radius)
#         self.samplePoints.append(l)

#         return l

#     def add_contours(self, levels=5, **kwargs):
#         """
#         Add constant centered RMS difference contours, defined by *levels*.
#         """

#         rs, ts = np.meshgrid(
#             np.linspace(self.xmin, self.xmax), np.linspace(0, self.tmax)
#         )
#         # Compute centered RMS difference
#         crmse = np.sqrt(self.refstd**2 + rs**2 - 2 * self.refstd * rs * np.cos(ts))
#         contours = self.ax.contour(ts, rs, crmse, levels, linestyles="--", **kwargs)
#         self.ax.clabel(contours, contours.levels[::1], inline=False)

#         return contours

#     def add_legend(self, *args, **kwargs):
#         return self.ax.legend(*args, **kwargs)

#     def add_annotation(self, *args, **kwargs):
#         return self.ax.annotation(*args, **kwargs)

#     def savefig(self, *args, **kwargs):
#         self.fig.savefig(*args, **kwargs)
    



import pathlib

import numpy as np
import pandas as pd
import xarray as xr

from scipy.interpolate import (
    RegularGridInterpolator,
    LinearNDInterpolator,
    NearestNDInterpolator,
)
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


class GridOperation:

    
    def lon_n2g(self, x):
        """
        Change normal longitude to GENIE longitude
        """
        if x > 100 and x < 180:
            return x - 360
        else:
            return x

    def lon_g2n(self, x):
        """
        Change GENIE longitude to normal longitude
        """

        if x < -180:
            return x + 360
        else:
            return x

    def lon_e2n(self, x):
        """
        Change eastern longitude to normal longitude
        """
        if x > 180:
            return x - 360
        else:
            return x

    def lon_n2e(self, x):
        """
        normal longitude (-180, 180) to longitude east(0,360)
        """
        if x < 0:
            return 360 + x
        else:
            return x

    def apply_n2g(self, data: xr.Dataset, longitude="lon") -> xr.Dataset:
        return data.assign_coords(
            {longitude: list(map(self.lon_n2g, data[longitude].values))}
        ).sortby("lon")

    def apply_g2n(self, data: xr.Dataset, longitude="lon") -> xr.Dataset:
        """
        Change GENIE longitude to normal longitude
        """
        return data.assign_coords(
            {longitude: list(map(self.lon_g2n, data[longitude].values))}
        ).sortby(longitude)

    def apply_e2n(self, data: xr.Dataset, longitude="lon") -> xr.Dataset:
        """
        Change eastern longitude to normal longitude
        """
        return data.assign_coords(
            {longitude: list(map(self.lon_e2n, data[longitude].values))}
        ).sortby(longitude)

    def apply_n2e(self, data: xr.Dataset, longitude="lon") -> xr.Dataset:
        """
        Change normal longitude to eastern longitude
        """
        return data.assign_coords(
            {longitude: list(map(self.lon_n2e, data[longitude].values))}
        ).sortby(longitude)

    def mask_Arctic_Med(self, array, policy="na"):
        """
        mask Arctic and Meditterean Sea in cGENIE modern continent configuration
        """
        if policy == "na":
            array[34:36, :] = np.nan
            array[27:30, 25:30] = np.nan
        elif policy == "zero":
            array[34:36, :] = 0
            array[27:30, 25:30] = 0

        return array

    def GENIE_grid_mask(
        self, base="worjh2", basin="ALL", subbasin="", mask_Arc_Med=False, invert=False
    ):
        """
        Get a modern GENIE 36x36 mask array from input data.
        The input array is flipped (left/right flip -> up/down flip) for easy recognition

        :continent: worjh2, worlg4, worbe2, GIteiiaa, GIteiiaa, p0055c
        :basin: Atlantic/Pacific/Indian/ALL/Tanzania
        :subbasin: N/S/ALL, ALL means Southern Ocean section included

        :returns: GENIE grid array where continent/ice cap is 0 and ocean is 1, default is 'worjh2'
        """

        file_path = (
            pathlib.Path(__file__).parent.parent
            / f"data/mask_{base}_{basin}{subbasin}.txt"
        )
        grid_mask_raw = np.loadtxt(file_path, dtype=int)
        grid_mask = np.flip(np.fliplr(grid_mask_raw))

        if invert:
            grid_mask = ~grid_mask + 2

        return grid_mask

    def normal_lon(self, N=36, edge=False):
        """
        Normal longitude in 10 degree resolution,
        if edge is False, then return midpoint
        """
        if edge:
            lon_edge = np.linspace(-180, 180, N + 1)
            return lon_edge
        else:
            lon = np.linspace(-175, 175, N)
            return lon

    def regrid_lat(self, x):

        """
        Transform <latitude> into cGENIE resolution to facilitate comparison between
        model and observational data.
        """
        if x >= -90 and x <= 90:
            lat_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, 37)))
            lat = np.rad2deg(np.arcsin(np.linspace(-1, 1, 36)))

            for i in range(36):
                if x > lat_edge[i] and x <= lat_edge[i + 1]:
                    x = lat[i]
        else:
            raise ValueError("Latitude must be in [-90,90]")

        return x

    def regrid_lon(self, x):

        """
        Transform <longitude> into cGENIE resolution to facilitate comparison between
        model and observational data.
        """

        if x >= -180 and x <= 180:
            lon_edge = np.linspace(-180, 180, 37)
            for i in range(36):
                if x > lon_edge[i] and x <= lon_edge[i + 1]:
                    x = (lon_edge[i] + lon_edge[i + 1]) / 2  # middle value in the bin
        else:
            raise ValueError("Longitude must be in [-180,180]")

        return x

    def haversine_distance(self, lat1, lon1, lat2, lon2):
        """
        Calculate the Haversine distance between corresponding pairs of points.

        :param lat1: Array of latitudes for points 1
        :param lon1: Array of longitudes for points 1
        :param lat2: Array of latitudes for points 2
        :param lon2: Array of longitudes for points 2
        
        :returns: Array of distances between corresponding pairs of points
        """
        # Convert latitude and longitude from degrees to radians
        lat1_rad, lon1_rad, lat2_rad, lon2_rad = np.radians(lat1), np.radians(lon1), np.radians(lat2), np.radians(lon2)

        # Haversine formula
        dlon = lon2_rad - lon1_rad
        dlat = lat2_rad - lat1_rad
        a = np.sin(dlat / 2) ** 2 + np.cos(lat1_rad) * np.cos(lat2_rad) * np.sin(dlon / 2) ** 2
        c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

        # Radius of the Earth in kilometers
        earth_radius = 6371.0
        distance = earth_radius * c

        return distance

    def geo_dis3d(self, point1, points2):
        """
        Calculate the 3D geographical distance between point1 and multiple points2.

        :param point1: Tuple/list of coordinates (z, lat, lon) or (lat, lon)
        :param points2: Array of shape (n, 3) containing coordinates (z, lat, lon) of points
        :returns: Array of distances between point1 and each point in points2
        """
        z1, lat1, lon1 = point1
        z2, lat2, lon2 = points2[:, 0], points2[:, 1], points2[:, 2]
        hor_dis = self.haversine_distance(lat1, lon1, lat2, lon2)
        ver_dis = np.abs(z1 - z2) / 1000  # Convert meters to kilometers
        return np.sqrt(hor_dis**2 + ver_dis**2)

    def geo_dis2d(self, point1, points2):
        """
        Calculate the 2D geographical distance between point1 and multiple points2.

        :param point1: Tuple/list of coordinates (z, lon, lat) or (lon, lat)
        :param points2: Array of shape (n, 3) containing coordinates (z, lon, lat) of points
        :returns: Array of distances between point1 and each point in points2
        """
        lat1, lon1 = point1[0], point1[1]
        lat2, lon2 = points2[:, 0], points2[:, 1]
        return self.haversine_distance(lat1, lon1, lat2, lon2)

    def check_dimension(self,input):
        # Convert tuple elements to lowercase
        input_lower = tuple(element.lower() for element in input)

        # Initialize flags for presence of each element
        has_lat = False
        has_lon = False
        has_depth = False
        has_time = False

        lat_candidates = ['lat', 'latitude', 'y']
        lon_candidates = ['lon', 'longitude', 'x']
        depth_candidates = ['depth', 'z', 'z_t', 'level', 'nlevel', 'lvl','lev', 'depth_1',
                            'elevation [m]']
        time_candidates = ['time', 't', 'age', 'date', 'year']

        # Check for presence of each element
        for element in input_lower:
            if element in lat_candidates:
                has_lat = True
            if element in lon_candidates:
                has_lon = True
            if element in depth_candidates:
                has_depth = True
            if element in time_candidates:
                has_time = True

        return has_lat, has_lon, has_depth, has_time

    def dim_order(self,input):
        """
        Determine the order of dimensions in the input data array

        return: tuple of index in the input data

        Example: input = ('lat', 'lon', 'depth')
                 return: (2, 0, 1) (always follow time, depth, lat, lon)
        """
        order = []
        input_lower = tuple(element.lower() for element in input)

        dim_candidates = [
            ['time', 't', 'age', 'date', 'year'],
            ['depth', 'z', 'z_t', 'level', 'nlevel', 'lvl', 'lev', 'depth_1',
                'elevation [m]'],
            ['lat', 'latitude', 'y'],
            ['lon', 'longitude', 'x']
        ]

        for candidates in dim_candidates:
            for candidate in candidates:
                if candidate in input_lower:
                    order.append(input_lower.index(candidate))
                    break

        return tuple(order)
    
    @staticmethod
    def set_coordinates(obj, index):
        obj.n_index = len(index)
        gp = GridOperation()
        has_lat, has_lon, has_depth, has_time = gp.check_dimension(index)
        index_order = gp.dim_order(index)
        match obj.n_index:
            case 1:
                if has_lat: obj.lat = index[0]
                if has_lon: obj.lon = index[0]                
                if has_depth: obj.depth = index[0]
                if has_time: obj.time = index[0]
            case 2:
                if has_lat and has_lon:
                    obj.lat = index[index_order[0]]
                    obj.lon = index[index_order[1]]
                if has_lat and has_depth:
                    obj.depth = index[index_order[0]]
                    obj.lat = index[index_order[1]]
                if has_lat and has_time:
                    obj.time = index[index_order[0]]
                    obj.lat = index[index_order[1]]
                if has_lon and has_depth:
                    obj.depth = index[index_order[0]]
                    obj.lon = index[index_order[1]]
                if has_lon and has_time:
                    obj.time = index[index_order[0]]
                    obj.lon = index[index_order[1]]
                if has_depth and has_time:
                    obj.time = index[index_order[0]]
                    obj.depth = index[index_order[1]]
            case 3:
                if not has_lat:
                    obj.time = index[index_order[0]]
                    obj.depth = index[index_order[1]]
                    obj.time = index[index_order[0]]                                          
                if not has_lon:
                    obj.time = index[index_order[0]]
                    obj.depth = index[index_order[1]]
                    obj.lat = index[index_order[2]]
                if not has_depth:
                    obj.time = index[index_order[0]]
                    obj.lat = index[index_order[1]]
                    obj.lon = index[index_order[2]]
                if not has_time:
                    obj.lat = index[index_order[0]]
                    obj.lon = index[index_order[1]]
                    obj.depth = index[index_order[2]]
            case 4:
                    obj.time = index[index_order[0]]
                    obj.depth = index[index_order[1]]
                    obj.lat = index[index_order[2]]
                    obj.lon = index[index_order[3]]             

class Interporaltor:
    
    def __init__(self, dims, coordinates, values, grid_number=200, method="r-linear"):
        """
        initialize regridder

        :param array: xr data array or dataframe
        :param grid_number: number of target grid points
        :param method: interpolation method, only linear is supported
        """
        self.dims = dims
        self.coords = coordinates
        self.values = values

        ## number of grid points
        self.grid_number = grid_number
        ## create new coordinates
        self.gridded_coord = self.new_coordinate(n=grid_number)
        ## create meshgrid
        self.meshgrid = self.new_meshgrid(*self.gridded_coord, indexing="ij")
        ## interpolation function
        self.interp_function = self._create_interp_function(method=method)
        ## interpolate data
        self.gridded_data = self.interpolate_data(tuple(self.meshgrid))

    def _create_interp_function(self, method):
        """
        create interpolation function

        :param method: interpolation method, use "x-y" format. where x is strcutre or not, y is the algorithm
        """

        data_class = method.split("-")[0]
        ## regular grid interpolation
        if data_class == "r":
            true_method = method.split("-")[1]
            interp_function = RegularGridInterpolator(
                self.coords, self.values, method=true_method
            )
        ## irregular grid
        elif data_class == "ir":
            true_method = method.split("-")[1]
            if true_method == "linear":
                interp_function = LinearNDInterpolator(self.coords, self.values)
            elif true_method == "nearest":
                interp_function = NearestNDInterpolator(self.coords, self.values)
        else:
            raise ValueError("Method not supported")

        return interp_function

    def new_coordinate(self, n):
        "create new coordinates for regridding"
        new_coords = []
        for coord_values in self.coords:
            min_val, max_val = np.min(coord_values), np.max(coord_values)
            new_values = np.linspace(min_val, max_val, n)
            new_coords.append(new_values)
        return new_coords

    def new_meshgrid(self, *args, **kwargs):
        "regrid to finer dimensions"
        return np.meshgrid(*args, **kwargs)

    def interpolate_data(self, *args, **kwargs):
        "Use the interpolation function to get regridded values"
        return self.interp_function(*args, **kwargs)

    def to_xarray(self):
        """
        convert numpy array to xr data array
        """
        return xr.DataArray(
            self.gridded_data, dims=self.dims, coords=self.gridded_coord
        )



import numpy as np
import xarray as xr
from scipy.stats import sem
import regionmask

from .grid import Interporaltor, GridOperation
from .plot import GriddedDataVis
from .chem import Chemistry
import cgeniepy.table as ct
from functools import cache


class GriddedData:
    """
    GriddedData is a class to store and compute GENIE netcdf data.

    It stores data in xarray.DataArray format, and provides optimalised methods for GENIE model output to compute statistics.    
    """
    
    def __init__(self, array=np.nan, mutable=False, attrs={}):
        """
        Initialise an instance of GriddedData

        Parameters
        ----------
        array : xarray.DataArray
        """
        
        # set data
        self.data = array
        self.mutable = mutable
        self.attrs = attrs

        ## formatting the unit
        if 'units' in self.attrs:
            self.attrs['units'] = Chemistry().format_unit(self.attrs['units'])

    def __getitem__(self, item):
        "make GriddedData subscriptable like xarray.DataArray"
        return self.data[item]

    def interpolate(self, *args, **kwargs):
        """
        Interpolate the GriddedData to a finer grid, mostly useful for GENIE plotting
        """
        
        coords = tuple([self.data[dim].values for dim in self.data.dims])
        values = self.data.values
        dims = self.data.dims
        
        if self.mutable:
            self.data = Interporaltor(dims, coords,values, *args, **kwargs).to_xarray()
            return self            
        else:
            return GriddedData(Interporaltor(dims, coords,values, *args, **kwargs).to_xarray())
        

    def sel(self, *args, **kwargs):
        """select grid points based on the given coordinates
        a wrapper to xarray `sel` method

        -----------
        Examples:
        -----------
        >>> model = GenieModel(path)
        >>> var = model.get_var(target_var)
        >>> var.search_grid(lon=XX, lat=XX, zt=XX, method='nearest')
        
        ## for loop over a data frame (df) with longitude, latitude, depth columns
        >>> modelvar = []
        >>> for i in range(len(df)):
        >>>     lat = df.latitude.iloc[i][0]
        >>>     lon = df.longitude[i].iloc[0]
        >>>     zt = df.depth[i].iloc[0]
        >>>     kwargs = {'lat': lat, 'lon':lon, 'zt': zt, 'method': 'nearest'}
        >>>     var = var.sel(**kwargs).values
        >>>     modelvar.append(var)        
        """

        if self.mutable:
            self.data = self.data.sel(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.sel(*args, **kwargs))


    def isel(self, *args, **kwargs):
        """select grid points based on the given index
        a wrapper to xarray `isel` method
        """
        if self.mutable:
            self.data = self.data.isel(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.isel(*args, **kwargs), mutable=False, attrs=self.attrs)
    
    def normalise_longitude(self, method='g2n', *args, **kwargs):
        """Normalise GENIE's longitude (eastern degree) to normal longitude (-180, 180)

        -----------
        Example
        -----------

        >>> Model = GenieModel("a path")
        >>> Model.get_var('abc').normalise_longitude()
        """
        gp = GridOperation()

        match method:
            case 'g2n':
                output = gp.apply_g2n(self.data, *args, **kwargs)
            case 'n2g':
                output = gp.apply_n2g(self.data, *args, **kwargs)
            case 'e2n':
                output = gp.apply_e2n(self.data, *args, **kwargs)
            case 'n2e':
                output = gp.apply_n2e(self.data, *args, **kwargs)
            case _:
                raise ValueError("Invalid method")

        if self.mutable:
            self.data = output
            return self        
        else:
             return GriddedData(output, mutable=False, attrs=self.attrs)
                    

    def __add__(self, other):
        """Allow GriddedData to be added by a number or another GriddedData
        """
        sum = GriddedData()
        if hasattr(other, "array"):
            sum.data = self.data + other.array
        else:
            ## a scalar
            sum.data = self.data + other
        return sum

    def __sub__(self, other):
        """
        Allow GriddedData to be subtracted by a number or another GriddedData
        """
        diff = GriddedData()
        if hasattr(other, "data"):
            diff.data = self.data - other.data
        else:
            ## a scalar
            diff.data = self.data - other
        return diff

    def __truediv__(self, other):
        """
        Allow GriddedData to be divided by a number or another GriddedData

        NA/NA -> NA
        """
        quotient_array = np.zeros_like(self.data)
        
        if isinstance(other, GriddedData):
            quotient_array = np.divide(self.data, other.data)
        else:
            # Handle division by a scalar
            quotient_array = np.divide(self.data, other)

        quotient = GriddedData()
        quotient.data = xr.DataArray(quotient_array)
        return quotient

    def __mul__(self, other):
        """
        Allow GriddedData to be multiplied by a number or another GriddedData
        """
        product = GriddedData()
        if hasattr(other, "data"):
            product.data = self.data * other.data
        else:
            try:
                product.data = self.data * other
            except ValueError:
                print("Sorry, only number and GriddedData are accepted")

        return product


    def max(self, *args, **kwargs):
        self.data = self.data.max(*args, **kwargs)
        return self


    def min(self, *args, **kwargs):
        if self.mutable:
            self.data = self.data.min(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.min(*args, **kwargs), mutable=False, attrs=self.attrs)


    def sum(self, *args, **kwargs):
        if self.mutable:
            self.data = self.data.sum(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.sum(*args, **kwargs), mutable=False, attrs=self.attrs)


    def mean(self, *args, **kwargs):
        if self.mutable:
            self.data = self.data.mean(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.mean(*args, **kwargs), mutable=False, attrs=self.attrs)
    

    def median(self, *args, **kwargs):
        if self.mutable:
            self.data = self.data.median(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.median(*args, **kwargs), mutable=False, attrs=self.attrs)

    def sd(self, *args, **kwargs):
        "standard deviation of the mean"
        if self.mutable:
            self.data = np.std(self.data, *args, **kwargs)
            return self
        else:
            return np.std(self.data, *args, **kwargs)


    def variance(self, *args, **kwargs):
        if self.mutable:
            self.data = np.var(self.data, *args, **kwargs)
            return self
        else:
            return np.var(self.data, *args, **kwargs)

    def se(self, *args, **kwargs):
        "standard error of the mean"
        if self.mutable:
            self.data = sem(self.data, nan_policy="omit", axis=None, *args, **kwargs)
            return self
        else:
            return sem(self.data, nan_policy="omit", axis=None, *args, **kwargs)

    def weighted_mean(self, weights, *args, **kwargs):
        """
        compute the weighted average of the array (e.g., ocean volume)
        """
        array_ma = self.data.to_masked_array()
        return np.ma.average(array_ma, weights=weights, *args, **kwargs)        
    
    def sel_modern_basin(self, basin):
        """
        select modern basin from regionmask.defined_regions.ar6.ocean

        The 58 defines marine regions from the sixth IPCC assessment report (AR6), Iturbide et al., (2020) ESSD
        
        https://regionmask.readthedocs.io/en/stable/_images/plotting_ar6_all.png
        
        --------------------------------
        47: North Pacific
        48: Equatorial Pacific
        49: South Pacific
        
        50: North Atlantic Ocean
        51: Equatorial Atlantic Ocean
        52: Southern Atlantic Ocean

        57: Southern Ocean
        --------------------------------        
        """
        ocean = regionmask.defined_regions.ar6.ocean
        index = ocean.map_keys(basin)
        mask = ocean.mask(self.normalise_longitude())        

        if self.mutable:
            self.data = self.normalise_longitude().data.where(mask == index)                    
            return self
        else:
            return GriddedData(self.normalise_longitude().data.where(mask == index), mutable=False, attrs=self.attrs)


    def mask_basin(self, base, basin, subbasin):

        """
        use pre-defined grid mask to select_basin, mostly used for cGENIE model
        """
        gp = GridOperation()
        data = self.data
        mask = gp.GENIE_grid_mask(base=base, basin=basin, subbasin=subbasin, invert=True)

        if self.data.ndim > 2:
            mask = np.broadcast_to(mask, (16, 36, 36))

        mask_data = np.ma.array(data, mask=mask)
        mask_data = np.ma.masked_invalid(mask_data)


        if self.mutable:
            self.data = xr.DataArray(mask_data, dims=data.dims, coords=data.coords)            
            return self
        else:
            return xr.DataArray(mask_data, dims=data.dims, coords=data.coords)


    @cache
    def ocn_only_data(self, index=False):
        """
        remove the NA grid (i.e., land definition)

        :param index: return GeoIndex if True or real value if false
        """

        stacked_data = self.data.stack(x=self.data.dims)            
        stacked_ocn_mask =~np.isnan(stacked_data)
        ocn_only_data = stacked_data[stacked_ocn_mask]
        
        if index:
            return np.stack(ocn_only_data.x.values)
        else:
            return ocn_only_data
        

    def search_point(self, point, ignore_na=False, to_genielon=False, **kwargs):
        """
        search the nearest grid point to the given coordinates
        
        :param point: a list/tuple of coordinate values in the same order to the data dimension (self.data.dims)
        :param ignore_na: whether only check ocean data (which ignore the NA grids)
        :param to_genielon: whether convert the input longitude to genie longitude, input point must be list if True
        """
        ## ignore the first dimension (which is time)        
        ndim = self.data.ndim
        
        if len(point) != ndim:
            raise ValueError("Input point has incompatiable coordinate")
        
        if ignore_na:
            index_pool = self.ocn_only_data(index=True)
            match ndim:
                case 3:
                    ## point: (depth, lat, lon); Index: (depth, lat, lon)
                    if to_genielon: point[2] = GridOperation().lon_g2n(point[2])
                    distances = GridOperation().geo_dis3d(point, index_pool[:, 0:4])
                    idx_min = np.argmin(distances)
                case 2:
                    ## point: (lat, lon); Index: (lat, lon)
                    if to_genielon: point[1] = GridOperation().lon_g2n(point[1])
                    distances = GridOperation().geo_dis2d(point, index_pool[:, 0:3])
                    idx_min = np.argmin(distances)
            
            nearest_value = self.ocn_only_data(index=False).values[idx_min]
            return nearest_value
        else:
            ## construct a dictionary based on self.data.dims and input point, and method ("nearest")
            kwargs = dict(zip(self.data.dims, point), method='nearest')
            return self.data.sel(**kwargs).values.item()

    def to_GriddedDataVis(self):
        """
        convert GriddedData to GriddedDataVis
        """
        return GriddedDataVis(self)

    def plot(self, *args, **kwargs):
        """
        plot the data using GriddedDataVis
        """
        return self.to_GriddedDataVis().plot(*args, **kwargs)

    def to_dataframe(self):
        """
        convert GriddedData to pandas.DataFrame
        """
        return self.data.to_dataframe()

    def to_ScatterData(self):
        """
        convert GriddedData to ScatterData
        """
        return ct.ScatterData(self.to_dataframe())
